#! /usr/bin/env python
# -*- coding: latin-1 -*-

"""
This draws the correlation between two images..
"""

import time
import os
import sys
import numpy
import Image
from pylab import *
from pylab import *
from scipy import *
from scipy.stats.stats import *
from optparse import OptionParser
from pyhdf import SD
from pyhdf.SD import SDC
from pyhdf.HDF import *
from pyhdf.VS import *

class Correlation:

    def __init__(self,logfile,plot=True,density=False,fdensity=False):
        """
        Constructor
        """
        self.axisMinMax=False
        self.plot = plot
        self.density = density 
        self.fdensity = fdensity
        self.logger=Logger(True,logfile)

    def SetAxis(self,xmin,xmax,ymin,ymax):
        """
        If a section of the axis is two be selected then call this method.
        @param xmin X axix minimum
        @param xmax X axix maximum
        @param ymin Y axix minimum
        @param ymax Y axix maximum
        """
        self.axisMinMax=True
        self.xmin=xmin
        self.xmax=xmax 
        self.ymin=ymin
        self.ymax=ymax

    def SetValidValues(self,xmin,xmax,ymin,ymax):
        """
        If a section of the axis is two be selected then call this method.
        @param xmin X axix minimum
        @param xmax X axix maximum
        @param ymin Y axix minimum
        @param ymax Y axix maximum
        """        
        self.xvmin=xmin
        self.xvmax=xmax 
        self.yvmin=ymin
        self.yvmax=ymax

    def HDF(self,filename1,sdsname1,filename2,sdsname2,plot='bo'):
        """
        correlates two HFD
        @param filename1 first file name.
        @param sdsname1 first sds name
        @param filename2 second file name.
        @param sdsname2 second sds name.        
        @param plot plot parameters.
        """

        data1=self.hdf2array(filename1,sdsname1)
        data2=self.hdf2array(filename2,sdsname2)

        [data1,data2]=self.correl(data1,data2)
        [data1,data2]=self.format(filename1,sdsname1,filename2,sdsname2,data1,data2)
        [data1,data2]=self.filter_Val(data1,data2)

        self.create(data1,data2,plot)       

        self.statistics(data1,data2)      

    def images(self, im1, min1, max1, im2, min2, max2,plot='bo'):
        """
        correlates two images
        @param im1 image 1
        @param min1 minimum palette image 1
        @param max1 maximum palette image 1
        @param im2 image 2
        @param min2 minimum palette image 2
        @param max2 maximum palette image 2
        """
        data1 = self.image2array(im1,min1,max1)
        data2 = self.image2array(im2,min2,max2)

        [data1,data2]=self.correl(data1,data2)
        [data1,data2]=self.filter_Val(data1,data2)

        self.create(data1,data2,plot)

        self.statistics(data1,data2)

    def create(self, data1, data2, _plot):
        """
        plotes all points
        @param data1 first image array.
        @param data2 second image array.
        """
        if(len(data1) != len(data2) ):
            raise ValueError, "arrays not compatible"
            exit()                  

        xmin = data1.min()
        xmax = data1.max()
        ymin = data2.min()
        ymax = data2.max()
        if(self.axisMinMax):
            if(xmin < self.xmin): 
                xmin = self.xmin
            if(xmax > self.xmax): 
                xmax = self.xmax
            if(ymin < self.ymin): 
                ymin = self.ymin
            if(ymax > self.ymax): 
                ymax = self.ymax                             
        grid(True)                          

        figure(figsize=(5, 5))

        axis([xmin,xmax,ymin,ymax])     

        if(self.fdensity):
            # Perform a kernel density estimator on the results                             
            n=50
            Z,edgesX,edgesY = numpy.histogram2d(data1,data2,(n,n),normed=True)
            Z.reshape((n,n))
            
            imshow(rot90(Z),aspect='auto',extent=[xmin, xmax, ymin, ymax],interpolation='nearest')

            self.densityMin= min(Z.flat)
            self.densityMax= max(Z.flat)
 

        elif(self.density):
            # Perform a kernel density estimator on the results           
            X, Y = mgrid[xmin:xmax:50j, ymin:ymax:50j]
            positions = c_[X.ravel(), Y.ravel()]
            values = c_[data1, data2]
            kernel = stats.kde.gaussian_kde(values.T)            
            Z = reshape(kernel(positions.T).T, X.T.shape)                           
            imshow(rot90(Z),aspect='auto',extent=[xmin, xmax, ymin, ymax])

            self.densityMin= min(Z.flat)
            self.densityMax= max(Z.flat)
        
        else :
            self.densityMin= None
            self.densityMax= None    



        if(self.plot):            
            plot(data1, data2, _plot)       

    def statistics(self,data1,data2):

        self.logger.log("Mean X : \t"+str(mean(data1)))

        self.logger.log("Mean Y : \t"+str(mean(data2)))

        self.logger.log("Std X : \t"+str(std(data1)))

        self.logger.log("Std Y : \t"+str(std(data2)))

        self.logger.log("Skewness X : \t"+str(skew(data1)))

        self.logger.log("Skewness Y : \t"+str(skew(data2)))

        self.logger.log("Kurtosis Fisher X : \t"+str(kurtosis(data1, fisher=True)))

        self.logger.log("Kurtosis Fisher Y : \t"+str(kurtosis(data2, fisher=True)))

        self.logger.log("Kurtosis Pearson X : \t"+str(kurtosis(data1, fisher=False)))

        self.logger.log("Kurtosis Pearson Y : \t"+str(kurtosis(data2, fisher=False)))

        self.logger.log("Correlation coefficients :\n" +str(corrcoef(data1,data2)))

        size = len(data1)
        M = numpy.empty((size,2))

        M[0:size,0]=data1
        M[0:size,1]=data2
    
        self.logger.log("Covariance :\n"+str(cov(M)))

        if(self.densityMin!=None and self.densityMax!= None):
            self.logger.log("Density Minimum :\t"+str(self.densityMin))        
            self.logger.log("Density Maximum :\t"+str(self.densityMax))

    def save(self,filename):
        """ 
        save result
        @param filename file name to save
        """
        savefig(filename)            

    def image2array(self,file, min0,max0):    
        """
        transforms a image to a one dimension array
        @param file file name
        @param min0 palette minimum.
        @param max0 palette maximum.
        """
        im = Image.open(file)          
        array = numpy.array(im.getdata())
        array.shape=im.size        
        inf = (max0-min0)/256.0  
        array = array*inf+min0     
        return array

    def hdf2array(self,filename,sdsname):
        """
        reads a HDF file and returns a array with the specified sds values.
        @param filename the file name.
        @param sdsname the name of the sds
        @return array of values.
        """
        hdf = SD.SD(filename,SDC.READ)
        sds = hdf.select(sdsname)
        array = sds.get()
        return array

    def correl(self,array1, array2):
        """
        correls two arrays having different number of values.
        @param array1 first array
        @param array2 second array
        """
        x1 = array1.shape[0]
        x2 = array2.shape[0]
        y1 = array1.shape[1]
        y2 = array2.shape[1]

        if x1 > x2 :
            xmax=x1
            if y1 > y2 :
                ymax=y1
                res1 = empty((x1,y1))
                res2 = empty((x1,y1))
                kx = float(x2)/float(x1)
                ky = float(y2)/float(y1)
                for i in range(x1):
                    for j in range(y1):
                        ii = int(i*kx)
                        jj = int(j*ky)
                        res1[i,j]=array1[i,j]
                        res2[i,j]=array2[ii,jj]
            else:
                ymax=y2
                res1 = empty((x1,y2))
                res2 = empty((x1,y2))
                kx = float(x2)/float(x1)
                ky = float(y1)/float(y2)
                for i in range(x1):
                    for j in range(y2):
                        ii = int(i*kx)
                        jj = int(j*ky)
                        res1[i,j]=array1[i,jj]
                        res2[i,j]=array2[ii,j]
        else :
            xmax=x2
            if y1 > y2 :
                ymax=y1
                res1 = empty((x2,y1))
                res2 = empty((x2,y1))
                kx = float(x1)/float(x2)
                ky = float(y2)/float(y1)
                for i in range(x2):
                    for j in range(y1):
                        ii = int(i*kx)
                        jj = int(j*ky)
                        res1[i,j]=array1[ii,j]
                        res2[i,j]=array2[i,jj]
            else:
                ymax=y2
                res1 = empty((x2,y2))
                res2 = empty((x2,y2))
                kx = float(x1)/float(x2)
                ky = float(y1)/float(y2)
                for i in range(x2):
                    for j in range(y2):
                        ii = int(i*kx)
                        jj = int(j*ky)
                        res1[i,j]=array1[ii,jj]
                        res2[i,j]=array2[i,j]
    
        return [res1.reshape(xmax*ymax,1),res2.reshape(xmax*ymax,1)]

    def format(self,filename1,sdsname1,filename2,sdsname2,data1,data2):
        """
        formats values with scale factors and offsets and eliminates fill values.
        """
        hdf1 = SD.SD(filename1,SDC.READ)
        sds1 = hdf1.select(sdsname1)
        attributes1 = sds1.attributes(False)     
        fillvalue1 = attributes1["_FillValue"]  
        scale1=attributes1["scale_factor"]
        offset1=attributes1["add_offset"]

        hdf2 = SD.SD(filename2,SDC.READ)
        sds2 = hdf2.select(sdsname2)
        attributes2 = sds2.attributes(False)     
        fillvalue2 = attributes2["_FillValue"]   
        scale2=attributes2["scale_factor"]       
        offset2=attributes2["add_offset"]
        
        mask = numpy.where(numpy.logical_and(data1!=fillvalue1,data2!=fillvalue2))        

        data1 = (data1[mask]-offset1)*scale1
        data2 = (data2[mask]-offset2)*scale2

        return [data1,data2]

    def filter_Val(self,data1, data2):
        if(self.xvmin!=""):
            xvmin = float(self.xvmin)
            mask = numpy.where(data1>=xvmin)        
            data1 = data1[mask]
            data2 = data2[mask]

        if(self.yvmin!=""):
            yvmin = float(self.yvmin)
            mask = numpy.where(data2>=yvmin)        
            data1 = data1[mask]
            data2 = data2[mask]

        if(self.xvmax!=""):
            xvmax = float(self.xvmax)
            mask = numpy.where(data1<=xvmax)        
            data1 = data1[mask]
            data2 = data2[mask]

        if(self.yvmax!=""):
            yvmax = float(self.yvmax)
            mask = numpy.where(data1<=yvmax)        
            data1 = data1[mask]
            data2 = data2[mask]    

        return [data1,data2]


class Logger():
    """logger"""

    def __init__(self, trace, filename):
        """Constructor"""
        self.trace = trace
        self.filename = filename       
        try:
            self.file = open(filename, 'w')
        except:
            self.error("Can't create logfile from : "+self.filename)

    def __del__(self):
        """Destructor"""        
        self.file.close()
     
    def log(self, msg):       
        """log a message"""
        if self.trace :
            print msg
        self.file.write(msg +'\n')
            
    def error(self, msg):
        """log a error and quits"""
        print msg
        self.file.write(msg +'\n')
        exit(1)





##############  MAIN  ###############

### set the available program options ###
def init_prog_options(parser):

    parser.add_option("--xamin","--xamin",
                      action="store", default="",
                      type="string",dest="xmin",help="the minimum of the X axis.")

    parser.add_option("--xamax","--xamax",
                      action="store", default="",
                      type="string",dest="xmax",help="the maximum of the X axis.")

    parser.add_option("--yamin","--yamin",
                      action="store", default="",
                      type="string",dest="ymin",help="the minimum of the Y axis.")

    parser.add_option("--yamax","--yamax",
                      action="store", default="",
                      type="string",dest="ymax",help="the maximum of the Y axis.")

    parser.add_option("--xvmin","--xvmin",
                      action="store", default="",
                      type="string",dest="xvmin",help="the minimum of the X values.")

    parser.add_option("--xvmax","--xvmax",
                      action="store", default="",
                      type="string",dest="xvmax",help="the maximum of the X values.")

    parser.add_option("--yvmin","--yvmin",
                      action="store", default="",
                      type="string",dest="yvmin",help="the minimum of the Y values.")

    parser.add_option("--yvmax","--yvmax",
                      action="store", default="",
                      type="string",dest="yvmax",help="the maximum of the Y values.")

    parser.add_option("--i1","--image1",
                      action="store", default="",
                      type="string",dest="image1",help="name of the directory and file of the first image.")
    parser.add_option("--i2","--image2",
                      action="store", default="",
                      type="string",dest="image2",help="name of the directory and file of the second image.")
    parser.add_option("--f1","--filename1",
                      action="store", default="",
                      type="string",dest="filename1",help="name of the directory and first HDF file.")
    parser.add_option("--f2","--filename2",
                      action="store", default="",
                      type="string",dest="filename2",help="name of the directory and second HDF file.")
    parser.add_option("--s1","--sds1",
                      action="store", default="",
                      type="string",dest="sds1",help="Name of the first SDS.")
    parser.add_option("--s2","--sds2",
                      action="store", default="",
                      type="string",dest="sds2",help="Name of the second SDS.")


    parser.add_option("--min1","--min1",
                      action="store", default="0.0",
                      type="float",dest="min1",help="minimum of the palette of image one.")
    parser.add_option("--min2","--min2",
                      action="store", default="0.0",
                      type="float",dest="min2",help="minimum of the palette of image two.")
    parser.add_option("--max1","--max1",
                      action="store", default="100.0",
                      type="float",dest="max1",help="maximum of the palette of image one.")
    parser.add_option("--max2","--max2",
                      action="store", default="100.0",
                      type="float",dest="max2",help="maximum of the palette of image two.")
    parser.add_option("-o","--output",
                      action="store", default="/tmp/output.png",
                      type="string",dest="outfile",help="The output file name.")
    parser.add_option("-s", "--small", action="store_false", dest="small", \
                      default=True,
                      help="Plot small dots.")
    parser.add_option("-c", "--col", action="store", dest="color", \
                      default="blue",
                      help="Color of the dots (red/blue/green/yellow/black).")
    parser.add_option("-p", "--plot", action="store_true", dest="dplot", \
                      default=False,
                      help="Plot dots.")
    parser.add_option("-d", "--density", action="store_true", dest="density", \
                      default=False,
                      help="Draws density (slow and hight resolution).")
    parser.add_option("-f", "--fdensity", action="store_true", dest="fdensity", \
                      default=False,
                      help="Draws density (fast and low resolution).")
    parser.add_option("--stat", "--stat", action="store", dest="logfile", \
                      default="/tmp/stat.txt",
                      help="Statistical output.")

def parse_arguments(parser):
    # parse command line
    return parser.parse_args()

def main():
    """
    correlates the two images (palettes values are supposed to be linear and are supposed to have 256 values)
    """        

    time_start=time.time()

    #---set the command-line options
    parser = OptionParser()
    init_prog_options(parser)
    #---parse them
    (options, args) = parse_arguments(parser)


    arg = sys.argv

    if(options.color == "red"): 
        plot='r'
    elif(options.color == "blue"):
        plot='b'    
    elif(options.color == "green"):
        plot='g'    
    elif(options.color == "yellow"):
        plot='y'    
    elif(options.color == "black"):
        plot='k'    
    else :
        plot='b'

    if(options.small): 
        plot+='o'
    else:
        plot+='.'
        
    if(not options.dplot and not options.density and not options.fdensity) :
        options.dplot=True

    x =  Correlation(options.logfile,options.dplot,options.density,options.fdensity)
    
    x.SetValidValues(options.xvmin,options.xvmax,options.yvmin,options.yvmax)

    if(options.xmin!="" and options.xmax!="" and options.ymin!="" and options.ymax!=""):
        x.SetAxis(float(options.xmin),float(options.xmax),float(options.ymin),float(options.ymax))

    if(options.filename1!=""):
        if(options.filename2==""):
            options.filename2=options.filename1        
        x.HDF(options.filename1,options.sds1,options.filename2,options.sds2,plot)
    else:
        x.images(options.image1,options.min1,options.max1,options.image2,options.min2,options.max2,plot)       

    x.save(options.outfile)

    time_end=time.time()
    print "Exec time : "+str(time_end-time_start)     

if __name__ == '__main__':    
    main()

